# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the `iris.fileformats.netcdf._load_aux_factory` function."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from unittest import mock

import numpy as np
import warnings

from iris.coords import DimCoord
from iris.cube import Cube
from iris.fileformats.netcdf import _load_aux_factory


class TestAtmosphereHybridSigmaPressureCoordinate(tests.IrisTest):
    def setUp(self):
        standard_name = "atmosphere_hybrid_sigma_pressure_coordinate"
        self.requires = dict(formula_type=standard_name)
        self.ap = mock.MagicMock(units="units")
        self.ps = mock.MagicMock(units="units")
        coordinates = [(mock.sentinel.b, "b"), (self.ps, "ps")]
        self.cube_parts = dict(coordinates=coordinates)
        self.engine = mock.Mock(
            requires=self.requires, cube_parts=self.cube_parts
        )
        self.cube = mock.create_autospec(Cube, spec_set=True, instance=True)
        # Patch out the check_dependencies functionality.
        func = "iris.aux_factory.HybridPressureFactory._check_dependencies"
        patcher = mock.patch(func)
        patcher.start()
        self.addCleanup(patcher.stop)

    def test_formula_terms_ap(self):
        self.cube_parts["coordinates"].append((self.ap, "ap"))
        self.requires["formula_terms"] = dict(ap="ap", b="b", ps="ps")
        _load_aux_factory(self.engine, self.cube)
        # Check cube.add_aux_coord method.
        self.assertEqual(self.cube.add_aux_coord.call_count, 0)
        # Check cube.add_aux_factory method.
        self.assertEqual(self.cube.add_aux_factory.call_count, 1)
        args, _ = self.cube.add_aux_factory.call_args
        self.assertEqual(len(args), 1)
        factory = args[0]
        self.assertEqual(factory.delta, self.ap)
        self.assertEqual(factory.sigma, mock.sentinel.b)
        self.assertEqual(factory.surface_air_pressure, self.ps)

    def test_formula_terms_a_p0(self):
        coord_a = DimCoord(np.arange(5), units="1")
        coord_p0 = DimCoord(10, units="Pa")
        coord_expected = DimCoord(
            np.arange(5) * 10,
            units="Pa",
            long_name="vertical pressure",
            var_name="ap",
        )
        self.cube_parts["coordinates"].extend(
            [(coord_a, "a"), (coord_p0, "p0")]
        )
        self.requires["formula_terms"] = dict(a="a", b="b", ps="ps", p0="p0")
        _load_aux_factory(self.engine, self.cube)
        # Check cube.coord_dims method.
        self.assertEqual(self.cube.coord_dims.call_count, 1)
        args, _ = self.cube.coord_dims.call_args
        self.assertEqual(len(args), 1)
        self.assertIs(args[0], coord_a)
        # Check cube.add_aux_coord method.
        self.assertEqual(self.cube.add_aux_coord.call_count, 1)
        args, _ = self.cube.add_aux_coord.call_args
        self.assertEqual(len(args), 2)
        self.assertEqual(args[0], coord_expected)
        self.assertIsInstance(args[1], mock.Mock)
        # Check cube.add_aux_factory method.
        self.assertEqual(self.cube.add_aux_factory.call_count, 1)
        args, _ = self.cube.add_aux_factory.call_args
        self.assertEqual(len(args), 1)
        factory = args[0]
        self.assertEqual(factory.delta, coord_expected)
        self.assertEqual(factory.sigma, mock.sentinel.b)
        self.assertEqual(factory.surface_air_pressure, self.ps)

    def test_formula_terms_a_p0__promote_a_units_unknown_to_dimensionless(
        self,
    ):
        coord_a = DimCoord(np.arange(5), units="unknown")
        coord_p0 = DimCoord(10, units="Pa")
        coord_expected = DimCoord(
            np.arange(5) * 10,
            units="Pa",
            long_name="vertical pressure",
            var_name="ap",
        )
        self.cube_parts["coordinates"].extend(
            [(coord_a, "a"), (coord_p0, "p0")]
        )
        self.requires["formula_terms"] = dict(a="a", b="b", ps="ps", p0="p0")
        _load_aux_factory(self.engine, self.cube)
        # Check cube.coord_dims method.
        self.assertEqual(self.cube.coord_dims.call_count, 1)
        args, _ = self.cube.coord_dims.call_args
        self.assertEqual(len(args), 1)
        self.assertIs(args[0], coord_a)
        self.assertEqual("1", args[0].units)
        # Check cube.add_aux_coord method.
        self.assertEqual(self.cube.add_aux_coord.call_count, 1)
        args, _ = self.cube.add_aux_coord.call_args
        self.assertEqual(len(args), 2)
        self.assertEqual(args[0], coord_expected)
        self.assertIsInstance(args[1], mock.Mock)
        # Check cube.add_aux_factory method.
        self.assertEqual(self.cube.add_aux_factory.call_count, 1)
        args, _ = self.cube.add_aux_factory.call_args
        self.assertEqual(len(args), 1)
        factory = args[0]
        self.assertEqual(factory.delta, coord_expected)
        self.assertEqual(factory.sigma, mock.sentinel.b)
        self.assertEqual(factory.surface_air_pressure, self.ps)

    def test_formula_terms_p0_non_scalar(self):
        coord_p0 = DimCoord(np.arange(5))
        self.cube_parts["coordinates"].append((coord_p0, "p0"))
        self.requires["formula_terms"] = dict(p0="p0")
        with self.assertRaises(ValueError):
            _load_aux_factory(self.engine, self.cube)

    def test_formula_terms_p0_bounded(self):
        coord_a = DimCoord(np.arange(5))
        coord_p0 = DimCoord(1, bounds=[0, 2], var_name="p0")
        self.cube_parts["coordinates"].extend(
            [(coord_a, "a"), (coord_p0, "p0")]
        )
        self.requires["formula_terms"] = dict(a="a", b="b", ps="ps", p0="p0")
        with warnings.catch_warnings(record=True) as warn:
            warnings.simplefilter("always")
            _load_aux_factory(self.engine, self.cube)
            self.assertEqual(len(warn), 1)
            msg = (
                "Ignoring atmosphere hybrid sigma pressure scalar "
                "coordinate {!r} bounds.".format(coord_p0.name())
            )
            self.assertEqual(msg, str(warn[0].message))

    def _check_no_delta(self):
        # Check cube.add_aux_coord method.
        self.assertEqual(self.cube.add_aux_coord.call_count, 0)
        # Check cube.add_aux_factory method.
        self.assertEqual(self.cube.add_aux_factory.call_count, 1)
        args, _ = self.cube.add_aux_factory.call_args
        self.assertEqual(len(args), 1)
        factory = args[0]
        # Check that the factory has no delta term
        self.assertEqual(factory.delta, None)
        self.assertEqual(factory.sigma, mock.sentinel.b)
        self.assertEqual(factory.surface_air_pressure, self.ps)

    def test_formula_terms_ap_missing_coords(self):
        self.requires["formula_terms"] = dict(ap="ap", b="b", ps="ps")
        with mock.patch("warnings.warn") as warn:
            _load_aux_factory(self.engine, self.cube)
        warn.assert_called_once_with(
            "Unable to find coordinate for variable " "'ap'"
        )
        self._check_no_delta()

    def test_formula_terms_no_delta_terms(self):
        self.requires["formula_terms"] = dict(b="b", ps="ps")
        _load_aux_factory(self.engine, self.cube)
        self._check_no_delta()

    def test_formula_terms_no_p0_term(self):
        coord_a = DimCoord(np.arange(5), units="Pa")
        self.cube_parts["coordinates"].append((coord_a, "a"))
        self.requires["formula_terms"] = dict(a="a", b="b", ps="ps")
        _load_aux_factory(self.engine, self.cube)
        self._check_no_delta()

    def test_formula_terms_no_a_term(self):
        coord_p0 = DimCoord(10, units="1")
        self.cube_parts["coordinates"].append((coord_p0, "p0"))
        self.requires["formula_terms"] = dict(a="p0", b="b", ps="ps")
        _load_aux_factory(self.engine, self.cube)
        self._check_no_delta()


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the :data:`iris.analysis.MEAN` aggregator."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from iris._lazy_data import as_lazy_data
import numpy as np
import numpy.ma as ma

from iris.analysis import MEAN
from iris._lazy_data import as_concrete_data


class Test_lazy_aggregate(tests.IrisTest):
    def setUp(self):
        self.data = ma.arange(12).reshape(3, 4)
        self.data.mask = [[0, 0, 0, 1], [0, 0, 1, 1], [0, 1, 1, 1]]
        # --> fractions of masked-points in columns = [0, 1/3, 2/3, 1]
        self.array = as_lazy_data(self.data)
        self.axis = 0
        self.expected_masked = ma.mean(self.data, axis=self.axis)

    def test_mdtol_default(self):
        # Default operation is "mdtol=1" --> unmasked if *any* valid points.
        # --> output column masks = [0, 0, 0, 1]
        agg = MEAN.lazy_aggregate(self.array, axis=self.axis)
        masked_result = as_concrete_data(agg)
        self.assertMaskedArrayAlmostEqual(masked_result, self.expected_masked)

    def test_mdtol_belowall(self):
        # Mdtol=0.25 --> masked columns = [0, 1, 1, 1]
        agg = MEAN.lazy_aggregate(self.array, axis=self.axis, mdtol=0.25)
        masked_result = as_concrete_data(agg)
        expected_masked = self.expected_masked
        expected_masked.mask = [False, True, True, True]
        self.assertMaskedArrayAlmostEqual(masked_result, expected_masked)

    def test_mdtol_intermediate(self):
        # mdtol=0.5 --> masked columns = [0, 0, 1, 1]
        agg = MEAN.lazy_aggregate(self.array, axis=self.axis, mdtol=0.5)
        masked_result = as_concrete_data(agg)
        expected_masked = self.expected_masked
        expected_masked.mask = [False, False, True, True]
        self.assertMaskedArrayAlmostEqual(masked_result, expected_masked)

    def test_mdtol_aboveall(self):
        # mdtol=0.75 --> masked columns = [0, 0, 0, 1]
        # In this case, effectively the same as mdtol=None.
        agg = MEAN.lazy_aggregate(self.array, axis=self.axis, mdtol=0.75)
        masked_result = as_concrete_data(agg)
        self.assertMaskedArrayAlmostEqual(masked_result, self.expected_masked)

    def test_multi_axis(self):
        data = np.arange(24.0).reshape((2, 3, 4))
        collapse_axes = (0, 2)
        lazy_data = as_lazy_data(data)
        agg = MEAN.lazy_aggregate(lazy_data, axis=collapse_axes)
        result = as_concrete_data(agg)
        expected = np.mean(data, axis=collapse_axes)
        self.assertArrayAllClose(result, expected)

    def test_last_axis(self):
        # From setUp:
        # self.data.mask = [[0, 0, 0, 1],
        #                   [0, 0, 1, 1],
        #                   [0, 1, 1, 1]]
        # --> fractions of masked-points in ROWS = [1/4, 1/2, 3/4]
        axis = -1
        agg = MEAN.lazy_aggregate(self.array, axis=axis, mdtol=0.51)
        expected_masked = ma.mean(self.data, axis=-1)
        expected_masked = np.ma.masked_array(expected_masked, [0, 0, 1])
        masked_result = as_concrete_data(agg)
        self.assertMaskedArrayAlmostEqual(masked_result, expected_masked)

    def test_all_axes_belowtol(self):
        agg = MEAN.lazy_aggregate(self.array, axis=None, mdtol=0.75)
        expected_masked = ma.mean(self.data)
        masked_result = as_concrete_data(agg)
        self.assertMaskedArrayAlmostEqual(masked_result, expected_masked)

    def test_all_axes_abovetol(self):
        agg = MEAN.lazy_aggregate(self.array, axis=None, mdtol=0.45)
        expected_masked = ma.masked_less([0.0], 1)
        masked_result = as_concrete_data(agg)
        self.assertMaskedArrayAlmostEqual(masked_result, expected_masked)


class Test_name(tests.IrisTest):
    def test(self):
        self.assertEqual(MEAN.name(), "mean")


class Test_aggregate_shape(tests.IrisTest):
    def test(self):
        shape = ()
        kwargs = dict()
        self.assertTupleEqual(MEAN.aggregate_shape(**kwargs), shape)
        kwargs = dict(one=1, two=2)
        self.assertTupleEqual(MEAN.aggregate_shape(**kwargs), shape)


if __name__ == "__main__":
    tests.main()
